#include "Particle.h"
#include "tga.h"

using namespace std;

ParticleManager::ParticleManager()
{
  int Loop;
  PStruct NewBit;

  Bits.reserve(PARTICLECOUNT);

  NewBit.Active=false;

  for(Loop=0;Loop!=PARTICLECOUNT;++Loop)
  {
    Bits.push_back(NewBit);
  }

  Current=0;
}

bool ParticleManager::LoadTexture(char* fName)
{
  TGAImg Img;        // Image loader

  // Load our Texture
  if(Img.Load(fName)!=IMG_OK)
    return false;

  glGenTextures(1,&ParticleTex);            // Allocate space for texture
  glBindTexture(GL_TEXTURE_2D,ParticleTex); // Set our Tex handle as current

  // Create the texture
  if(Img.GetBPP()==24)
    glTexImage2D(GL_TEXTURE_2D,0,3,Img.GetWidth(),Img.GetHeight(),0,GL_RGB,GL_UNSIGNED_BYTE,Img.GetImg());
  else if(Img.GetBPP()==32)
    glTexImage2D(GL_TEXTURE_2D,0,4,Img.GetWidth(),Img.GetHeight(),0,GL_RGBA,GL_UNSIGNED_BYTE,Img.GetImg());
  else
    return false;

  // Specify filtering and edge actions
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);

  return true;
}

// Add a new particle
void ParticleManager::Add(float x, float y, float z, float xv, float yv, float zv)
{
  if(Current==PARTICLECOUNT)
    Current=0;

  Bits[Current].x=x;
  Bits[Current].y=y;
  Bits[Current].z=z;
  Bits[Current].xv=xv;
  Bits[Current].yv=yv;
  Bits[Current].zv=zv;
  Bits[Current].Active=true;
  Bits[Current].Type=rand()%2;

  ++Current;
}


// Update and draw all active particles.
void ParticleManager::Render(double TBMVal)
{
  int Loop;
  vector<PStruct>::iterator BitPtr;

  glEnable(GL_BLEND);
  glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA);
  glBindTexture(GL_TEXTURE_2D,ParticleTex);
  glDisable(GL_LIGHTING);
  glDepthMask(false);

  //BitPtr=Bits.begin();

  for(Loop=0;Loop!=PARTICLECOUNT;++Loop)
  {
    if(Bits[Loop].Active)
    {
      // Add motion vector to position
      Bits[Loop].x+=Bits[Loop].xv*TBMVal;
      Bits[Loop].y+=Bits[Loop].yv*TBMVal;
      Bits[Loop].z+=Bits[Loop].zv*TBMVal;

      // Add a little gravity
      Bits[Loop].yv-=0.000005*TBMVal;

      // Kill particles that have fallen beyond the bottom of the screen
      if(Bits[Loop].y<-10.0f)
        Bits[Loop].Active=false;

      // Draw 0s or 1s
      if(Bits[Loop].Type)
      {
        glBegin(GL_QUADS);
        glTexCoord2f(0.0f,0.5f); glVertex3f(Bits[Loop].x-PARTICLESIZE,Bits[Loop].y-PARTICLESIZE,Bits[Loop].z);
        glTexCoord2f(1.0f,0.5f); glVertex3f(Bits[Loop].x+PARTICLESIZE,Bits[Loop].y-PARTICLESIZE,Bits[Loop].z);
        glTexCoord2f(1.0f,0.0f); glVertex3f(Bits[Loop].x+PARTICLESIZE,Bits[Loop].y+PARTICLESIZE,Bits[Loop].z);
        glTexCoord2f(0.0f,0.0f); glVertex3f(Bits[Loop].x-PARTICLESIZE,Bits[Loop].y+PARTICLESIZE,Bits[Loop].z);
        glEnd();
      }
      else
      {
        glBegin(GL_QUADS);
        glTexCoord2f(0.0f,1.0f); glVertex3f(Bits[Loop].x-PARTICLESIZE,Bits[Loop].y-PARTICLESIZE,Bits[Loop].z);
        glTexCoord2f(1.0f,1.0f); glVertex3f(Bits[Loop].x+PARTICLESIZE,Bits[Loop].y-PARTICLESIZE,Bits[Loop].z);
        glTexCoord2f(1.0f,0.5f); glVertex3f(Bits[Loop].x+PARTICLESIZE,Bits[Loop].y+PARTICLESIZE,Bits[Loop].z);
        glTexCoord2f(0.0f,0.5f); glVertex3f(Bits[Loop].x-PARTICLESIZE,Bits[Loop].y+PARTICLESIZE,Bits[Loop].z);
        glEnd();
      }
    }
  }
  glDisable(GL_BLEND);
  glDepthMask(true);
}

