// This code will be executed by the OpenGL base app's render thread
#include "Framework.h"
#include "tga.h"
#include "BitmapFontClass.h"
#include "Particle.h"
#include <iostream>
#include <string>
#include <sstream>
#include <ctime>
#include <cstdlib>
#include <cmath>

#define START_RAM 4000000
#define SPIN_SPEED 0.15f

typedef struct VectorStruct
{
  float x,y,z;
}
Vec;

CBitmapFont Fnt; // CBFG Font
int Best;

GLuint LoadTexture(char *TexName);
Vec RotateVector(float X, float Y, float Z, double* Matrix);
int CreateLeak(ParticleStruct* Leak);
void DrawSelect();
void TexRender(GLuint Src, GLuint Tgt, GLuint Decal, int Location, bool Jitter);
void StartInfo(void);
bool GameOver(int Score);


using namespace std;

void Render(void)
{
  ParticleManager Pts;
  GLuint FrontBackTex,SideTex,HoleTex,PatchTex; // Texture handles
  GLuint RightTex,LeftTex,EndTex; // Dynamic texture handles

  double StartTime,Score,CurTime,TgtTime,LastFrame,HoldTime,TBMVal; // Timing vars

  ParticleStruct Emitter[12]; // Misuse of particle structure to hold emitter information
  float ScreenRatio;

  float Rot=60.0f; // Compybox initial rotation value

  long RAM;  
  stringstream Mem,Uptime;
  int MouseY;
  int Loop,FirstFree;
  char ColDex;
  int LeakID,TexSplat=0;
   
  double modelView[16]; // Array to hold modelview matrix
  Vec TransHole,TransEmit; // Transformed vectors

  

  // Set the background to black
  glClearColor(0,0,0,0);
  glClear(GL_COLOR_BUFFER_BIT);

  // Set up view
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glViewport(0,0,800,600);
  ScreenRatio=600.0f/800.0f;
  glFrustum(-1.0f,1.0f,-1.0f*ScreenRatio,1.0f*ScreenRatio,0.1f,500.0f);
  glMatrixMode(GL_MODELVIEW);

  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_TEXTURE_2D);

   // Load media
   if(!Fnt.Load("LeakFont.bff"))
    {
     cout<<"Unable to load 'LeakFont.bff'\n";
     return;
    }
    
   FrontBackTex=LoadTexture("Case.tga");
    if(FrontBackTex==-1)
     {
     cout<<"Unable to load 'Case.tga'\n";
     return;
    }
    
  SideTex=LoadTexture("CaseSide.tga");
   if(SideTex==-1)
    {
     cout<<"Unable to load 'CaseSide.tga'\n";
     return;
    }
    
  PatchTex=LoadTexture("Patch.tga");
   if(PatchTex==-1)
    {
     cout<<"Unable to load 'Patch.tga'\n";
     return;
    }
    
  HoleTex=LoadTexture("Hole.tga");
   if(HoleTex==-1)
    {
     cout<<"Unable to load 'Hole.tga'\n";
     return;
    }  
    
   if(!Pts.LoadTexture("Bits.tga"))
    {
     cout<<"Unable to load 'Bits.tga'\n";
     return;
    }  

  // Cheap way of initalising texture handles
  EndTex=LoadTexture("Case.tga");
  LeftTex=LoadTexture("CaseSide.tga");
  RightTex=LoadTexture("CaseSide.tga");

  // Init particle emitters
  for(Loop=0;Loop!=12;++Loop)
  {
    Emitter[Loop].Active=false;
    Emitter[Loop].Source=0;
  }

  // Show the user some instructions.
  StartInfo();
  
  // Randomise random things
  srand(time(NULL));

  // Set up a few flag values
  HoldTime=-1;
  CurTime=TgtTime=0;
  StartTime=glfwGetTime();
  RAM=START_RAM;
  Score=-1;
  Best=0;

  // This loop will run until Esc is pressed
  while(RunLevel)
  {
    if(Keys[VK_ESCAPE]) // Esc Key
      RunLevel=0;

    // Get time values for this frame
    LastFrame=CurTime;
    CurTime=glfwGetTime();
    TBMVal=(CurTime-LastFrame)*1000.0f;

    // Clear last frame
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Set up view
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(-1.0f,1.0f,-1.0f*ScreenRatio,1.0f*ScreenRatio,1.0f,100.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // Move to view position
    glTranslatef(0.0f,0.0f,-2.75f);

    // Save current modelview
    glPushMatrix();

    // Handle keyboard inputs
    if(Keys[GLFW_KEY_LEFT])
      Rot+=(SPIN_SPEED*TBMVal);

    if(Keys[GLFW_KEY_RIGHT])
      Rot-=(SPIN_SPEED*TBMVal);
 
    // Set compybox rotation
    glRotatef(Rot,0.0f,1.0f,0.0f);

    // Do cheapo selection if mouse button down
    if(Mouse.Mleft)
    {
      DrawSelect();
      glReadPixels(Mouse.Mx,600-Mouse.My,1,1,GL_RED,GL_BYTE,&ColDex);

      if(ColDex!=0)
      {
        for(Loop=0;Loop!=12;++Loop)
        {
          // Apply patch texture and stop emitter 
          // if user clicked on an active leak
          if(Emitter[Loop].Active && Emitter[Loop].Source==ColDex)
          {
            Emitter[Loop].Active=false;
            TexSplat=ColDex;
          }
        }
      }

      // Erase selection image
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }

    // Draw compybox
    glEnable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
    glEnable(GL_DEPTH_TEST);
    glColor3f(1.0f,1.0f,1.0f);

    // Right Side
    glBindTexture(GL_TEXTURE_2D,RightTex);
    glBegin(GL_QUADS);
    glNormal3f(0.0f,0.0f,-1.0f);
    glTexCoord2f(0.0f,1.0f); glVertex3f(-1.0f,-1.0f,-0.5f);
    glTexCoord2f(1.0f,1.0f); glVertex3f( 1.0f,-1.0f,-0.5f);
    glTexCoord2f(1.0f,0.0f); glVertex3f( 1.0f, 1.0f,-0.5f);
    glTexCoord2f(0.0f,0.0f); glVertex3f(-1.0f, 1.0f,-0.5f);
    glEnd();
    
    // Left Side
    glBindTexture(GL_TEXTURE_2D,LeftTex);
    glBegin(GL_QUADS);
    glNormal3f(0.0f,0.0f,1.0f);
    glTexCoord2f(0.0f,1.0f); glVertex3f(-1.0f,-1.0f, 0.5f);
    glTexCoord2f(1.0f,1.0f); glVertex3f( 1.0f,-1.0f, 0.5f);
    glTexCoord2f(1.0f,0.0f); glVertex3f( 1.0f, 1.0f, 0.5f);
    glTexCoord2f(0.0f,0.0f); glVertex3f(-1.0f, 1.0f, 0.5f);
    glEnd();

    // Back
    glBindTexture(GL_TEXTURE_2D,FrontBackTex);
    glBegin(GL_QUADS);
    glNormal3f(1.0f,0.0f,0.0f);
    glTexCoord2f(0.0f,1.0f); glVertex3f( 1.0f,-1.0f, 0.5f);
    glTexCoord2f(0.5f,1.0f); glVertex3f( 1.0f,-1.0f,-0.5f);
    glTexCoord2f(0.5f,0.0f); glVertex3f( 1.0f, 1.0f,-0.5f);
    glTexCoord2f(0.0f,0.0f); glVertex3f( 1.0f, 1.0f, 0.5f);
    
    // Front
    glNormal3f(-1.0f,0.0f,0.0f);
    glTexCoord2f(0.5f,1.0f); glVertex3f(-1.0f,-1.0f, 0.5f);
    glTexCoord2f(1.0f,1.0f); glVertex3f(-1.0f,-1.0f,-0.5f);
    glTexCoord2f(1.0f,0.0f); glVertex3f(-1.0f, 1.0f,-0.5f);
    glTexCoord2f(0.5f,0.0f); glVertex3f(-1.0f, 1.0f, 0.5f);
    glEnd();

    // Restore modelview (Cancel rotation)
    glPopMatrix();

    // Create modelview for calculation of emitter locations
    glPushMatrix();
    glLoadIdentity();
    glRotatef(Rot,0.0f,1.0f,0.0f);

    // Grab modelview matrix
    glGetDoublev(GL_MODELVIEW_MATRIX, modelView);
    
    // Restore normal view
    glPopMatrix();

    // Draw particles
    Pts.Render(TBMVal);

    // Draw GUI
    glDisable(GL_DEPTH_TEST);
    Fnt.SetScreen(800,600); // Cheap way of getting ortho screen

    // Draw patch at mouse location
    glBindTexture(GL_TEXTURE_2D,PatchTex);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    MouseY=600-Mouse.My;
    glBegin(GL_QUADS);
    glTexCoord2f(0.0f,1.0f); glVertex2i(Mouse.Mx-32,MouseY-32);
    glTexCoord2f(1.0f,1.0f); glVertex2i(Mouse.Mx+32,MouseY-32);
    glTexCoord2f(1.0f,0.0f); glVertex2i(Mouse.Mx+32,MouseY+32);
    glTexCoord2f(0.0f,0.0f); glVertex2i(Mouse.Mx-32,MouseY+32);
    glEnd();

    // Show status messages
    Fnt.Select();
    Mem.str("");
    if(RAM>0)
      Mem<<"Remaining RAM: "<<RAM<<"k";
    else
      Mem<<"Remaining RAM: 0";
    Uptime.str("");
    if(Score==-1)
      Uptime<<"Uptime: "<<(int)(CurTime-StartTime)<<" Seconds";
    else
      Uptime<<"Uptime: "<<(int)Score<<" Seconds";
    Fnt.ezPrint((char*)Mem.str().c_str(),20,30);
    Fnt.ezPrint((char*)Uptime.str().c_str(),600,30);


    // We're using double buffers, so we need to swap to see our stuff
    FlipBuffers();

    // Add particles
    if(TgtTime<CurTime && RAM>0)
    {
      // See if there is an unused emitter
      FirstFree=-1;
      for(Loop=0;Loop!=12;++Loop)
      {
        if(Emitter[Loop].Active==false)
        {
          FirstFree=Loop;
          break;
        }
      }

      // Create a random leak
      if(FirstFree>-1 && (rand()%50)==1)
      {
        LeakID=CreateLeak(&Emitter[FirstFree]);

        // Place a 'hole' on the texture
        if(LeakID)
        {
          if(LeakID<37)
          {
            TexRender(RightTex,RightTex,HoleTex,LeakID-1,false);
          }
          else if(LeakID<73)
          {
            TexRender(LeftTex,LeftTex,HoleTex,LeakID-37,false);
          }
          else
          {
            TexRender(FrontBackTex,FrontBackTex,HoleTex,LeakID-73,false);
          }
          LeakID=0;
        }
      }

      // Update particles
      for(Loop=0;Loop!=12;++Loop)
      {
        if(Emitter[Loop].Active)
        {
          TransEmit=RotateVector(Emitter[Loop].x,Emitter[Loop].y,Emitter[Loop].z,modelView);
          TransHole=RotateVector(Emitter[Loop].xv,Emitter[Loop].yv,Emitter[Loop].zv,modelView);
          Pts.Add(TransEmit.x,
                  TransEmit.y,
                  TransEmit.z,
                  TransHole.x+((rand()%100)-50)/100000.0f,
                  TransHole.y+(rand()%100)/400000.0f,
                  TransHole.z+((rand()%100)-50)/100000.0f);
          RAM-=(256*TBMVal); // Reduce memory total
        }
      }
      TgtTime=CurTime+0.015f; // Specify when next update will happen
    }

    // Place patch textures if required
    if(TexSplat)
    {
      if(TexSplat<37)
      {
        TexRender(RightTex,RightTex,PatchTex,TexSplat-1,true);
      }
      else if(TexSplat<73)
      {
        TexRender(LeftTex,LeftTex,PatchTex,TexSplat-37,true);
      }
      else
      {
        TexRender(FrontBackTex,FrontBackTex,PatchTex,TexSplat-73,true);
      }
      TexSplat=0;
    }

    // Check for game over
    if(RAM<1)
    {
      RAM=0;

      // Set a small delay to allow the user 
      // to realise that the game ha finished
      if(HoldTime==-1)
      {
        HoldTime=CurTime+2.0f;
        Score=(CurTime-StartTime);
      }
      else if(CurTime>HoldTime)
      {
        // Display end message and see if we're gonna play again
        if(GameOver(Score))
        {
          // Reset Emitters
          for(Loop=0;Loop!=12;++Loop)
            Emitter[Loop].Active=false;

          // Reset Textures
          TexRender(EndTex,FrontBackTex,PatchTex,TexSplat-73,true);
          TexRender(SideTex,LeftTex,PatchTex,TexSplat-73,true);
          TexRender(SideTex,RightTex,PatchTex,TexSplat-73,true);

          // Times
          StartTime=glfwGetTime();          
          HoldTime=-1;
          Score=-1;
          RAM=START_RAM;          
        }
        else
        {
          RunLevel=0;
        }
        glClearColor(0.0f,0.0f,0.0f,0.0f);
      }
    }
  }
}


// Texture loader
GLuint LoadTexture(char *TexName)
{
  TGAImg Img;        // Image loader
  GLuint Texture;

  // Load our Texture
  if(Img.Load(TexName)!=IMG_OK)
    return -1;

  glGenTextures(1,&Texture);            // Allocate space for texture
  glBindTexture(GL_TEXTURE_2D,Texture); // Set our Tex handle as current

  // Create the texture
  if(Img.GetBPP()==24)
    glTexImage2D(GL_TEXTURE_2D,0,3,Img.GetWidth(),Img.GetHeight(),0,GL_RGB,GL_UNSIGNED_BYTE,Img.GetImg());
  else if(Img.GetBPP()==32)
    glTexImage2D(GL_TEXTURE_2D,0,4,Img.GetWidth(),Img.GetHeight(),0,GL_RGBA,GL_UNSIGNED_BYTE,Img.GetImg());
  else
    return -1;

  // Specify filtering and edge actions
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);

  return Texture;
}

// Rotate emitter location and direction vectors
Vec RotateVector(float X, float Y, float Z, double* Matrix)
{
  Vec NewVec;

  NewVec.x=(Matrix[0]*X)+(Matrix[4]*Y)+(Matrix[8]*Z)+Matrix[12];
  NewVec.y=(Matrix[1]*X)+(Matrix[5]*Y)+(Matrix[9]*Z)+Matrix[13];
  NewVec.z=(Matrix[2]*X)+(Matrix[6]*Y)+(Matrix[10]*Z)+Matrix[14];

  return NewVec;
}

// Create random emiiter locations and direction vectors
// The emitters can be placed on any of the four sides.
// The sides have either 36 or 18 possible emitter locations.
int CreateLeak(ParticleStruct* Leak)
{
  int Direction=rand()%4;
  int Row,Col;

  switch(Direction)
  {
  case 0: // Front
    Row=rand()%6;
    Col=rand()%3;
    Leak->x=-1.0f;
    Leak->y=-1.0f+(Row*0.3334f)+0.1667f;
    Leak->z=-0.5f+(Col*0.3334f)+0.1667f;
    Leak->xv=-0.0025f;
    Leak->yv= 0.0001f;
    Leak->zv= 0.0f;
    Leak->Active=true;
    Leak->Source=(Col*6)+Row+73;
    break;

  case 1: // Left Side
    Row=rand()%6;
    Col=rand()%6;
    Leak->x=1.0f-(Col*0.3334f)-0.1667f;
    Leak->y=-1.0f+(Row*0.3334f)+0.1667f;
    Leak->z=0.5f;
    Leak->xv=0.0f;
    Leak->yv=0.0001f;
    Leak->zv=0.0025f;
    Leak->Active=true;
    Leak->Source=(Col*6)+Row+37;
    break;

  case 2: // Back
    Row=rand()%6;
    Col=rand()%3;
    Leak->x=1.0f;
    Leak->y=-1.0f+(Row*0.3334f)+0.1667f;
    Leak->z=-0.5f+(Col*0.3334f)+0.1667f;
    Leak->xv=0.0025f;
    Leak->yv=0.0001f;
    Leak->zv=0.0f;
    Leak->Active=true;
    Leak->Source=(Col*6)+Row+91;
    break;

  case 3: // Right Side
    Row=rand()%6;
    Col=rand()%6;
    Leak->x=1.0f-(Col*0.3334f)-0.1667f;
    Leak->y=-1.0f+(Row*0.3334f)+0.1667f;
    Leak->z=-0.5f;
    Leak->xv=0.0f;
    Leak->yv=0.0001f;
    Leak->zv=-0.0025f;
    Leak->Active=true;
    Leak->Source=(Col*6)+Row+1;
    break;
    
  default:
    Leak->Source=0;
  }

  return Leak->Source;
}

// Cheap selection using color values.
// A grid of quads within the four sides of the compbox.
// There are 36 quads on the sides of the case and 18 on the front and back
void DrawSelect()
{
  char RedVal;
  int Row,Col;
  float RowStart,ColStart;

  glDisable(GL_LIGHTING);
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_BLEND);
  glEnable(GL_DEPTH_TEST);

  RedVal=1;

  // Right Side
  RowStart=1.0f;
  for(Col=0;Col!=6;++Col)
  {
    ColStart=-1.0f;
    for(Row=0;Row!=6;++Row)
    {
      glColor3b(RedVal,0,0);
      glBegin(GL_QUADS);
      glVertex3f(RowStart,ColStart,-0.5f);
      glVertex3f(RowStart-0.3334f,ColStart,-0.5f);
      glVertex3f(RowStart-0.3334f,ColStart+0.3334f,-0.5f);
      glVertex3f(RowStart,ColStart+0.3334f,-0.5f);
      glEnd();

      ColStart+=0.3334f;
      ++RedVal;
    }
    RowStart-=0.3334f;
  }

  // Left Side
  RowStart=1.0f;
  for(Col=0;Col!=6;++Col)
  {
    ColStart=-1.0f;
    for(Row=0;Row!=6;++Row)
    {
      glColor3b(RedVal,0,0);
      glBegin(GL_QUADS);
      glVertex3f(RowStart,ColStart,0.5f);
      glVertex3f(RowStart-0.3334f,ColStart,0.5f);
      glVertex3f(RowStart-0.3334f,ColStart+0.3334f,0.5f);
      glVertex3f(RowStart,ColStart+0.3334f,0.5f);
      glEnd();

      ColStart+=0.3334f;
      ++RedVal;
    }
    RowStart-=0.3334f;
  }

  // Front
  RowStart=-0.5f;
  for(Col=0;Col!=3;++Col)
  {
    ColStart=-1.0f;
    for(Row=0;Row!=6;++Row)
    {
      glColor3b(RedVal,0,0);
      glBegin(GL_QUADS);
      glVertex3f(-1.0f,ColStart,RowStart);
      glVertex3f(-1.0f,ColStart,RowStart+0.3334f);
      glVertex3f(-1.0f,ColStart+0.3334f,RowStart+0.3334f);
      glVertex3f(-1.0f,ColStart+0.3334f,RowStart);
      glEnd();

      ColStart+=0.3334f;
      ++RedVal;
    }
    RowStart+=0.3334f;
  }

  // Back
  RowStart=-0.5f;
  for(Col=0;Col!=3;++Col)
  {
    ColStart=-1.0f;
    for(Row=0;Row!=6;++Row)
    {
      glColor3b(RedVal,0,0);
      glBegin(GL_QUADS);
      glVertex3f(1.0f,ColStart,RowStart);
      glVertex3f(1.0f,ColStart,RowStart+0.3334f);
      glVertex3f(1.0f,ColStart+0.3334f,RowStart+0.3334f);
      glVertex3f(1.0f,ColStart+0.3334f,RowStart);
      glEnd();

      ColStart+=0.3334f;
      ++RedVal;
    }
    RowStart+=0.3334f;
  }
}

// Place decals on texture maps
// The location value is derived from the color index returned 
// from the Select function.
void TexRender(GLuint Src, GLuint Tgt, GLuint Decal, int Location, bool Jitter)
{
  int x,y;

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0f,800.0f,0.0f,600.0f,-1.0f,100.0f);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glClear(GL_COLOR_BUFFER_BIT);
  glEnable(GL_TEXTURE_2D);
  glDisable(GL_LIGHTING);
  glDisable(GL_BLEND);
  glDisable(GL_DEPTH_TEST);

  // Calculate the position of the decal
  x=Location/6;
  y=(Location-(x*6));

  x=234-(int)(x*42.5f);
  y=234-(int)(y*42.5f);

  // Jitter if required to stop things looking too uniform
  if(Jitter)
  {
    x+=(rand()%8)-4;
    y+=(rand()%8)-4;
  }

  glBindTexture(GL_TEXTURE_2D,Src);

  // Prevent blurring of textures
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);

  // Draw the source texture
  glBegin(GL_QUADS);
  glTexCoord2i(0,0); glVertex2i(0,0);
  glTexCoord2i(1,0); glVertex2i(256,0);
  glTexCoord2i(1,1); glVertex2i(256,256);
  glTexCoord2i(0,1); glVertex2i(0,256);
  glEnd();

  // Restore texture filtering
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);

  // Bind the decal
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glBindTexture(GL_TEXTURE_2D,Decal);

  // Move to the required position
  glTranslatef(x,y,0);

  // Rotate if required for some variety
  if(Jitter)
  {
    glRotatef(rand()%90,0.0f,0.0f,1.0f);
  }

  // Draw the decal
  glBegin(GL_QUADS);
  glTexCoord2i(0,0); glVertex2i(-20,-20);
  glTexCoord2i(1,0); glVertex2i(20,-20);
  glTexCoord2i(1,1); glVertex2i(20,20);
  glTexCoord2i(0,1); glVertex2i(-20,20);
  glEnd();

  // Pick up the modified texture
  glBindTexture(GL_TEXTURE_2D,Tgt);
  glCopyTexImage2D(GL_TEXTURE_2D,0,GL_RGB,0,0,256,256,0);
}


// Splash screen
void StartInfo()
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0f,800.0f,0.0f,600.0f,-1.0f,100.0f);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glClearColor(0.0f,0.0f,0.0f,0.0f);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  Fnt.ezPrint("Memory Leak",320,500);
  Fnt.ezPrint("An entry for GPWiki.org's 3rd Coding Competition",150,450);
  Fnt.ezPrint("Your Mission: Stop leaky applications from using up all your RAM",80,380);
  Fnt.ezPrint("Use the Left/Right cursor keys to rotate the compybox,",120,300);
  Fnt.ezPrint("Left mouse button applies a patch to stop leaks.",160,270);

  Fnt.ezPrint("Press the Spacebar to start",260,100);

  FlipBuffers();

  // Wait for user to continue or quit
  while(!Keys[GLFW_KEY_SPACE])
  {
    glfwPollEvents();

    if(Keys[GLFW_KEY_ESC] || !RunLevel)
    {
      RunLevel=0;
      break;
    }

    if (glfwGetWindowParam(GLFW_OPENED) != GL_TRUE)
    {
      RunLevel=0;
      break;
    }
  }
}

// Exit/Restart screen
bool GameOver(int Score)
{
  stringstream FinalScore;

   if(Score>Best)
    Best=Score;

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0f,800.0f,0.0f,600.0f,-1.0f,100.0f);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glClearColor(0.0f,0.0f,0.75f,0.0f);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  Fnt.Select();
  Fnt.Print("GAME OVER!",325,520);
  Fnt.Print("zOMG! Your Operating System is teh suX0rz!",200,450);
  Fnt.Print("All your memory are leaked out.",250,400);
  FinalScore.str("");
  FinalScore<<"You managed to keep the system up for "<<Score<<" seconds.";
  Fnt.Print((char*)FinalScore.str().c_str(),160,350);
  FinalScore.str("");
  FinalScore<<"The best uptime this session is "<<Best<<" seconds";;
  Fnt.Print((char*)FinalScore.str().c_str(),210,320);
  Fnt.Print("Here are some important looking numbers that no-one understands.",80,280);
  Fnt.Print("FF00:A0C8 ca80:FFF0 FC12:DC10 DEAD:BEEF 0ACE:FACE",110,240);
  Fnt.Print("F00D:CACE ACDC:0187 AC26:EC56 BEEF:FACE 78D1:CF51",110,215);
  Fnt.Print("FC80:DE52 8FF0:06F2 BCF1:747D C0DE:4EAD EF51:A7D2",110,190);

  Fnt.Print("Press the Spacebar to play again, or hit Esc to go vote for this entry.",80,100);

  FlipBuffers();

  // Wait for user to continue or quit
  while(!Keys[VK_SPACE])
  {
    glfwPollEvents();

    if(Keys[GLFW_KEY_ESC] || !RunLevel)
      return false;

    if (glfwGetWindowParam(GLFW_OPENED) != GL_TRUE)
      return false;
  }

  return true;
}
